local name, private = ...

local list = private.objects.list
local dict = private.objects.dict
local L = private.L

private.objects.data = {}

------------------------------------------------------------
-------------------------- Links ---------------------------
------------------------------------------------------------

private.links = { byItem = dict(), byProject = dict(), unknownProjects = dict() } -- To be populated on load

function private.links:getProject(itemName)
    return self.byItem[itemName] or itemName
end

function private.links:getItem(projectName)
    return self.byProject[projectName] or projectName
end

function private.links:addLink(itemName, projectName)
    self.byItem:set(itemName, projectName)
    self.byProject:set(projectName, itemName)
    self.unknownProjects:unset(projectName)
end

function private.links:removeLink(linkName)
    -- Looks for an item or project by the given name, removing it if found (items are searched first).
    local itemName, projectName = ""
    
    -- Find the item and project name to unlink.
    if self.byItem:get(linkName) then
        itemName = linkName
        projectName = self.byItem:get(linkName)
    elseif self.byProject:get(linkName) then
        projectName = linkName
        itemName = self.byProject:get(linkName)
    else
        return
    end
    
    -- Remove the two from the link lists.
    self.byItem:unset(itemName)
    self.byProject:unset(projectName)
    
    -- Add the project back into unknown projects.
    local project = private.projectList:get(projectName)
    self:addUnknownProject(projectName, unpack(project))
    
    -- Reset the completion count for the artifact.
    local race = private.raceList:getRace(project[1])
    local artifact = race:getArtifact(itemName)
    if artifact then artifact.completed = 0 end
end

function private.links:addUnknownProject(projectName, raceID, rarity, icon)
    if not self.unknownProjects:get(projectName) then
        self.unknownProjects:set(projectName, { raceID, rarity, icon })
        private:debug("Unknown project name:", projectName)
    end
end

---[[
for k, v in pairs(private.baseLinks) do
    private.links:addLink(k, v)
end
--]]

------------------------------------------------------------
------------------------- Artifact -------------------------
------------------------------------------------------------

private.objects.data.artifact = private.objects.base()

function private.objects.data.artifact:init(race, rarity, itemID)
    -- Base values
    self.name = ""
    self.projectName = ""
    self.rarity = "common"
    self.completed = 0
    self.race = 0
    self.itemID = 0
    self.itemLink = ""
    self.projectIcon = ""
    self.itemIcon = ""
    
    self.race = race
    self.rarity = rarity
    self.itemID = itemID
    
    -- Get the item's info based on the ID
    local name, link, _, _, _, _, _, _, _, icon = GetItemInfo(itemID)
    self.name = name
    self.itemLink = link
    self.itemIcon = icon
    
    -- If there's an item-project link, get the project's name
    self.projectName = private.links:getProject(self.name)
end

function private.objects.data.artifact:serialise()
    -- Returns a table with the necessary artifact information
    -- This object is never deserialised; this is only intended
    -- for exporting data.
    local r = {}
    
    r.itemID = self.itemID
    r.rarity = self.rarity
    r.race = self.race
    r.completed = self.completed
    
    return r
end

function private.objects.data.artifact.__eq(t1, t2)
    -- Checks whether two artifact objects are the same artifact
    -- The item ID should be sufficient for this
    return t1.itemID == t2.itemID
end

------------------------------------------------------------
--------------------------- Race ---------------------------
------------------------------------------------------------

-- Race list
private.raceList = {}
private.projectList = dict()

function private.raceList:setRace(raceID, raceObject)
    self[raceID] = raceObject
end

function private.raceList:getRace(raceID)
    return self[raceID]
end

private.objects.data.race = private.objects.base()

function private.objects.data.race:init(raceID)
    -- Base values
    self.name = ""
    self.raceID = 0
    self.artifacts = { common = dict(), rare = dict(), pristine = dict() }
    self.completed = { common = 0, rare = 0, pristine = 0, total = 0 }
    self.numArtifacts = { common = 0, rare = 0, pristine = 0, total = 0 }
    self.totalCompleted = 0
    self.regularFragments = 0
    self.currentArtifact = {
        artifact = "",
        fragments = 50,
        sockets = 0,
        name = "",
    }
    self.keystoneID = 0
    self.numKeystones = 0
    self.keystoneFragments = 0
    self.totalFragments = 0
    
    -- Record the race ID itself
    self.raceID = raceID
    
    -- Get the basic info
    local name, _, keystoneID, numColl, numReq = GetArchaeologyRaceInfo(raceID)
    self.name = name
    self.keystoneID = keystoneID
    self.regularFragments = numColl
    self.currentArtifact.fragments = numReq
    
    private.raceList:setRace(raceID, self)
    
    self:load()
    self:update()
end

test = false
function private.objects.data.race:update()
    -- Get the number of available keystones
    self.numKeystones = GetItemCount(self.keystoneID)
    
    -- Select this race's artifact
    SetSelectedArtifact(self.raceID)
    
    -- Socket all available keystones
    while SocketItemToArtifact() do end
    
    -- Get information about the current artifact
    local name, _, _, _, _, keystones = GetActiveArtifactByRace(self.raceID)
    if not name then return end -- In case the player has not seen this race yet
    local itemName = private.links:getItem(name)
    self.currentArtifact.artifact = self:getArtifact(itemName)
    self.currentArtifact.sockets = keystones
    self.currentArtifact.name = name
    
    -- Count the number of fragments available and required
    local reg, key, req = GetArtifactProgress()
    self.regularFragments = reg
    self.keystoneFragments = key
    self.currentArtifact.fragments = req
    self.totalFragments = reg + key
    
    -- Load all information about completed artifacts
    local num = GetNumArtifactsByRace(self.raceID)
    self.completed.common = 0
    self.completed.rare = 0
    self.totalCompleted = 0
    for i = 1, num do
        local name, _, rarity, icon, _, _, _, _, completed = GetArtifactInfoByRace(self.raceID, i)
        private.projectList:set(name, { self.raceID, rarity, icon }) -- Memory leak here, if this runs too often. Fix if necessary.
        local itemName = private.links:getItem(name)
        local artifact = self:getArtifact(itemName)
        if not artifact then
            -- At this point, itemName is actually the project name
            private.links:addUnknownProject(itemName, self.raceID, rarity, icon)
            artifact = private.objects.data.artifact()
            artifact:init(self.raceID, -1, 0) -- This won't (shouldn't) show anywhere, but will keep the addon from crashing.
        end
        completed = max(artifact.completed, completed) -- Because the game will return 0 for the current artifact, even if that's not true
        artifact.completed = completed
        artifact.projectIcon = icon
        self.totalCompleted = self.totalCompleted + completed
        if completed > 0 then
            if rarity == 0 then self.completed.common = self.completed.common + 1 end
            if rarity == 1 then self.completed.rare = self.completed.rare + 1 end
        end
    end
    self.completed.total = self.completed.common + self.completed.rare
end

function private.objects.data.race:load()
    -- Reset the counters
    self.numArtifacts.common = 0
    self.numArtifacts.rare = 0
    self.numArtifacts.total = 0
    
    -- Load all information about all artifacts
    for i=1, #private.artifacts[self.raceID].common do
        local itemID = private.artifacts[self.raceID].common[i]
        local artifact = private.objects.data.artifact:new()
        artifact:init(self.raceID, "common", itemID)
        if not self.artifacts.common:get(artifact.name) then self.artifacts.common:set(artifact.name, artifact) end -- Don't add the artifact if it's already there!
        self.numArtifacts.common = self.numArtifacts.common + 1
        
        if private.artifacts[self.raceID].commonToPristine then
            pristineID = private.artifacts[self.raceID].commonToPristine[itemID]
            if pristineID then
                artifact.pristineItemID = pristineID
            end
        end
    end
    for i=1, #private.artifacts[self.raceID].rare do
        local itemID = private.artifacts[self.raceID].rare[i]
        local artifact = private.objects.data.artifact:new()
        artifact:init(self.raceID, "rare", itemID)
        if not self.artifacts.rare:get(artifact.name) then self.artifacts.rare:set(artifact.name, artifact) end -- Don't add the artifact if it's already there!
        self.numArtifacts.rare = self.numArtifacts.rare + 1
    end
    if private.artifacts[self.raceID].pristine then
        for i=1, #private.artifacts[self.raceID].pristine do
            -- Changed - there's now separate data for pristine artifacts and the quests they start.
            local itemID = private.artifacts[self.raceID].pristine[i]
            local questID = private.artifacts[self.raceID].pristineQuests[itemID]
            self.artifacts.pristine:set(itemID, questID) -- Common and rare use the artifact data as value, pristine simply use questID.
        end
        self.numArtifacts.pristine = #private.artifacts[self.raceID].pristine
    end
    
    self.numArtifacts.total = self.numArtifacts.common + self.numArtifacts.rare
end

function private.objects.data.race:getArtifact(itemName)
    return self.artifacts.common:get(itemName) or self.artifacts.rare:get(itemName)
end

function private.objects.data.race:Solve()
    -- Select the artifact,
    SetSelectedArtifact(self.raceID)
    -- socket all available keystones,
    while SocketItemToArtifact() do end
    -- then solve it
    SolveArtifact()
end

------------------------------------------------------------
-------------------- Achievement (base) --------------------
------------------------------------------------------------

private.objects.data.achievement = private.objects.base()
-- Put in the default return value text for when an achievement has been completed.
private.objects.data.achievement.completedLines = list()
private.objects.data.achievement.completedLines:add(L("This achievement has been completed!"))

function private.objects.data.achievement:init(achievementID)
    self.achievementID = achievementID
    self.lines = { L("No information available.") }
    self.completed = false
    
    local _, name = GetAchievementInfo(achievementID)
    self.name = name
end

function private.objects.data.achievement:getNumericProgress()
    -- Override for achievement type specific handling
    -- Returns: Two values, first the current progress, and then
    -- the maximum.
    return 0, 1
end

function private.objects.data.achievement:getProgress()
    -- Returns: A value between 0 and 1, corresponding to the completion percentage.
    -- The default function will return a number based on the results of getNumericProgress().
    -- For different behaviour, override.
    local count, max = self:getNumericProgress()
    if max == 0 then return 0 end -- Avoid division by 0
    return min(count / max, 1)
end

function private.objects.data.achievement:getLines()
    -- Override for achievement type specific handling
    -- Returns: The lines of text to display
    if self.completed then return self.completedLines end
    return self.lines
end

function private.objects.data.achievement:SetFrame(frame)
    self.frame = frame
end

------------------------------------------------------------
--------------- Achievement (criteria list) ----------------
------------------------------------------------------------

private.objects.data.achievement.criteriaList = private.objects.data.achievement()

function private.objects.data.achievement.criteriaList:init(achievementID)
    private.objects.data.achievement.init(self, achievementID)
    
    self.lines = list()
    self.numCriteria = GetAchievementNumCriteria(self.achievementID)
    self.firstLine = L("You still need to complete the following criteria:")
end

function private.objects.data.achievement.criteriaList:getNumericProgress()
    if self.completed then return self.numCriteria, self.numCriteria end
    -- Count the number of completed criteria
    local count = 0
    for i = 1, self.numCriteria do
        local _, _, completed = GetAchievementCriteriaInfo(self.achievementID, i)
        if completed then count = count + 1 end
    end
    
    if count == self.numCriteria then self.completed = true end
    
    return count, self.numCriteria
end

function private.objects.data.achievement.criteriaList:getLines()
    if self.completed then return self.completedLines end
    
    self.lines:clear()
    self.lines:add(self.firstLine)
    for i = 1, self.numCriteria do
        local name, _, completed = GetAchievementCriteriaInfo(self.achievementID, i)
        if not completed then self.lines:add(name) end
    end
    
    return self.lines
end

function private.objects.data.achievement.criteriaList:setFirstLine(str)
    -- Sets the first 'line' to display in the tooltip list
    self.firstLine = str
end

------------------------------------------------------------
--------------- Achievement (artifact list) ----------------
------------------------------------------------------------

private.objects.data.achievement.artifactList = private.objects.data.achievement()

function private.objects.data.achievement.artifactList:init(achievementID)
    private.objects.data.achievement.init(self, achievementID)
    
    self.artifacts = dict()
    self.numArtifacts = 0
    self.lines = list()
    
    self:RegisterEvent("BAG_UPDATE")
end

function private.objects.data.achievement.artifactList:addArtifact(raceID, itemID)
    local itemName = GetItemInfo(itemID)
    --local raceObject = private.raceList:getRace(raceID)
    self.artifacts:set(itemName, raceID)
    self.numArtifacts = self.numArtifacts + 1
    self.completed = false
end

function private.objects.data.achievement.artifactList:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    self.lines:add(L("You still need these artifacts:"))
    for itemName, raceID in pairs(self.artifacts) do
        local raceObject = private.raceList:getRace(raceID)
        local artifact = raceObject:getArtifact(itemName)
        if artifact.completed == 0 then
            self.lines:add(artifact.itemLink .. " (" .. raceObject.name .. ")")
        end
    end
    return self.lines
end

function private.objects.data.achievement.artifactList:getNumericProgress()
    if self.completed then return self.numArtifacts, self.numArtifacts end
    -- Counts how many of the artifacts have been solved
    local count = 0
    for itemName, raceID in pairs(self.artifacts) do
        local raceObject = private.raceList:getRace(raceID)
        local artifact = raceObject:getArtifact(itemName)
        if artifact.completed > 0 then count = count + 1 end
    end
    if count == self.numArtifacts then self.completed = true end
    return count, self.numArtifacts
end

function private.objects.data.achievement.artifactList:BAG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
--------------- Achievement (artifact count) ---------------
------------------------------------------------------------

private.objects.data.achievement.artifactCount = private.objects.data.achievement()

function private.objects.data.achievement.artifactCount:init(achievementID)
    private.objects.data.achievement.init(self, achievementID)
    
    self.artifactName = ""
    self.raceID = 0
    self.count = 0
    self.lines = list()
    
    self:RegisterEvent("BAG_UPDATE")
end

function private.objects.data.achievement.artifactCount:setArtifact(raceID, itemID, count)
    local itemName = GetItemInfo(itemID)
    --local raceObject = private.raceList:getRace(raceID)
    self.artifactName = itemName
    self.raceID = raceID
    self.count = count
    self.completed = false
end

function private.objects.data.achievement.artifactCount:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    --self.lines:add(L("You still need these artifacts:"))
    local raceObject = private.raceList:getRace(self.raceID)
    local artifact = raceObject:getArtifact(self.artifactName)
    --if artifact.completed == 0 then
    --    self.lines:add(artifact.itemLink .. " (" .. raceObject.name .. ")")
    --end
    self.lines:add(format(L("You need to solve %s %d more times."), artifact.name, self.count - artifact.completed))
    return self.lines
end

function private.objects.data.achievement.artifactCount:getNumericProgress()
    if self.completed then return self.count, self.count end
    -- Counts how many of the artifacts have been solved
    --local count = 0
    --for itemName, raceID in pairs(self.artifacts) do
    --    local raceObject = private.raceList:getRace(raceID)
    --    local artifact = raceObject:getArtifact(itemName)
    --    if artifact.completed > 0 then count = count + 1 end
    --end
    --if count == self.numArtifacts then self.completed = true end
    --return count, self.numArtifacts
    local raceObject = private.raceList:getRace(self.raceID)
    local artifact = raceObject:getArtifact(self.artifactName)
    if artifact.completed >= self.count then self.completed = true end
    return artifact.completed, self.count
end

function private.objects.data.achievement.artifactCount:BAG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
------------------- Achievement (rares) --------------------
------------------------------------------------------------

private.objects.data.achievement.rareCount = private.objects.data.achievement()

function private.objects.data.achievement.rareCount:init(achievementID, max)
    private.objects.data.achievement.init(self, achievementID)
    
    self.lines = list()
    self.max = max
    
    self:RegisterEvent("BAG_UPDATE")
end

function private.objects.data.achievement.rareCount:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    self.lines:add(format(L("You still need %d rare artifacts."), self.max - self:getCount()))
    return self.lines
end

function private.objects.data.achievement.rareCount:getCount()
    local count = 0
    for raceID, raceObject in pairs(private.raceList) do
        if type(raceID) == "number" then
            for k, v in pairs(raceObject.artifacts.rare) do
                if v.completed > 0 then count = count + 1 end
            end
        end
    end
    return count
end

function private.objects.data.achievement.rareCount:getNumericProgress()
    if self.completed then return self.max, self.max end
    if self:getCount() >= self.max then self.completed = true end
    return self:getCount(), self.max
end

function private.objects.data.achievement.rareCount:BAG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
------------------ Achievement (commons) -------------------
------------------------------------------------------------

private.objects.data.achievement.commonCount = private.objects.data.achievement()

function private.objects.data.achievement.commonCount:init(achievementID, max)
    private.objects.data.achievement.init(self, achievementID)
    
    self.lines = list()
    self.max = max
    
    self:RegisterEvent("BAG_UPDATE")
end

function private.objects.data.achievement.commonCount:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    self.lines:add(format(L("You still need %d common artifacts."), self.max - self:getCount()))
    return self.lines
end

function private.objects.data.achievement.commonCount:getCount()
    local count = 0
    for raceID, raceObject in pairs(private.raceList) do
        if type(raceID) == "number" then
            for k, v in pairs(raceObject.artifacts.common) do
                if v.completed > 0 then count = count + 1 end
            end
        end
    end
    return count
end

function private.objects.data.achievement.commonCount:getNumericProgress()
    if self.completed then return self.max, self.max end
    if self:getCount() >= self.max then self.completed = true end
    return self:getCount(), self.max
end

function private.objects.data.achievement.commonCount:BAG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
------------------- Achievement (skill) --------------------
------------------------------------------------------------

private.objects.data.achievement.skill = private.objects.data.achievement()

function private.objects.data.achievement.skill:init(achievementID, max)
    private.objects.data.achievement.init(self, achievementID)
    
    self.lines = list()
    self.max = max
    
    self:RegisterEvent("CHAT_MSG_SKILL")
end

function private.objects.data.achievement.skill:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    local _, _, archIndex = GetProfessions()
    if not archIndex then
        self.lines:add(L("You have not yet learned archaeology."))
        return self.lines
    end
    local _, _, skill = GetProfessionInfo(archIndex)
    self.lines:add(format(L("You still need %d skill points."), self.max - skill))
    return self.lines
end

function private.objects.data.achievement.skill:getNumericProgress()
    if self.completed then return self.max, self.max end
    -- Get the current skill level
    local _, _, archIndex = GetProfessions()
    if not archIndex then return 0, 0 end
    local _, _, skill = GetProfessionInfo(archIndex)
    
    if skill >= self.max then self.completed = true end
    return skill, self.max
end

function private.objects.data.achievement.skill:CHAT_MSG_SKILL()
    -- Update the achievement frame
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
----------------- Achievement (race count) -----------------
------------------------------------------------------------

private.objects.data.achievement.raceCount = private.objects.data.achievement()

function private.objects.data.achievement.raceCount:init(achievementID, max)
    private.objects.data.achievement.init(self, achievementID)
    
    self.lines = list()
    self.max = max
    
    self:RegisterEvent("BAG_UPDATE")
end

function private.objects.data.achievement.raceCount:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    self.lines:add(L("You still need to complete an artifact from these races:"))
    
    for raceID, raceObject in pairs(private.raceList) do
        if type(raceID) == "number" then
            if raceObject.totalCompleted == 0 then
                self.lines:add(raceObject.name)
            end
        end
    end
    
    return self.lines
end

function private.objects.data.achievement.raceCount:getNumericProgress()
    if self.completed then return self.max, self.max end
    local count = 0
    
    for raceID, raceObject in pairs(private.raceList) do
        if type(raceID) == "number" then
            if raceObject.totalCompleted > 0 then count = count + 1 end
        end
    end
    
    if count >= self.max then self.completed = true end
    return count, self.max
end

function private.objects.data.achievement.raceCount:BAG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end

------------------------------------------------------------
------------------ Achievement (pristine) ------------------
------------------------------------------------------------

private.objects.data.achievement.pristine = private.objects.data.achievement()

function private.objects.data.achievement.pristine:init(achievementID)
    private.objects.data.achievement.init(self, achievementID)
    
    self.quests = dict()
    self.numQuests = 0
    self.lines = list()
    
    self:RegisterEvent("QUEST_LOG_UPDATE")
end

function private.objects.data.achievement.pristine:addQuest(questID)
    -- Do the same magic trick as in race to get the quest's name
    GameTooltip:SetOwner(UIParent, "ANCHOR_CURSOR")
    GameTooltip:SetHyperlink("quest:" .. questID .. ":1")
    GameTooltip:Show()
    local questName = GameTooltipTextLeft1:GetText()
    GameTooltip:Hide()
    
    self.quests:set(questName, questID)
    self.numQuests = self.numQuests + 1
end

function private.objects.data.achievement.pristine:getLines()
    if self.completed then return self.completedLines end
    self.lines:clear()
    self.lines:add(L("You still need these pristine artifacts:"))
    
    for questName, questID in pairs(self.quests) do
        local completed = IsQuestFlaggedCompleted(questID)
        if not completed then
            self.lines:add(questName)
        end
    end
    
    return self.lines
end

function private.objects.data.achievement.pristine:getNumericProgress()
    if self.completed then return self.numQuests, self.numQuests end
    -- Counts how many of the artifacts have been solved
    --local count = 0
    --for itemName, raceID in pairs(self.artifacts) do
    --    local raceObject = private.raceList:getRace(raceID)
    --    local artifact = raceObject:getArtifact(itemName)
    --    if artifact.completed > 0 then count = count + 1 end
    --end
    --if count == self.numArtifacts then self.completed = true end
    
    local count = 0
    for questName, questID in pairs(self.quests) do
        local completed = IsQuestFlaggedCompleted(questID)
        if completed then
            count = count + 1
        end
    end
    
    return count, self.numQuests
end

function private.objects.data.achievement.pristine:QUEST_LOG_UPDATE()
    if self.frame then
        self.frame:update()
    end
end